@@ -85,7 +85,7 @@ class DerivationTree:
 
         if tree is None:
             tree = DerivationTree('<start>')
-        tree.__dict__.update(from_dict(next(ijson.items(json_str, ''))).__dict__)
+        tree.__dict__.update(from_dict(next(ijson.items(json_str.encode('utf-8'), ''))).__dict__)
 
         return tree
 
@@ -432,7 +432,7 @@ class DerivationTree:
 
     def __len__(self):
         if self.__len is None:
-            self.__len = sum([len(child) for child in (self.children or [])]) + 1
+            self.__len = len(self.paths())
 
         return self.__len
 
