@@ -21,7 +21,7 @@ from grammar_graph.gg import path_to_string
 
 import isla.derivation_tree
 from isla.fuzzer import GrammarCoverageFuzzer
-from isla.helpers import tree_to_string
+from isla.helpers import tree_to_string, assertions_activated
 from isla.solver import ISLaSolver
 from isla.type_defs import Grammar, ParseTree
 
@@ -306,6 +306,84 @@ class Evaluator:
         avg_inp_length: Dict[str, float] = {}
         median_inp_length: Dict[str, float] = {}
 
+        # The custom TAR parser is inconsistent with the TAR grammar for the expansion of the `<content>`
+        # nonterminal. This has been fixed by now. For existing evaluation data, we patch the k-paths
+        # that are affected by the issue.
+        tar_kpath_fix_map: Dict[str, Set[str]] = {
+            '<content> <content>-choice-1 <characters> <characters>-choice-1 <character>': {
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+                '<maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <character>',
+            },
+            '<content> <content>-choice-1 <characters> <characters>-choice-1 <characters>': {
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+                '<maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <characters>',
+            },
+            '<content> <content>-choice-1 <nuls> <nuls>-choice-1 <NUL>': {
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+                '<maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <NUL>',
+            },
+            '<content> <content>-choice-1 <nuls> <nuls>-choice-1 <nuls>': {
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+                '<maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls>',
+            },
+            '<entry> <entry>-choice-1 <content> <content>-choice-1 <characters>': {
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters>',
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+            },
+            '<entry> <entry>-choice-1 <content> <content>-choice-1 <nuls>': {
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls>',
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+            },
+            '<content> <content>-choice-1 <characters> <characters>-choice-1 <characters> <characters>-choice-1 <character>': {
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <characters>',
+                '<maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <characters> <characters>-choice-1 <character>',
+            },
+            '<content> <content>-choice-1 <characters> <characters>-choice-1 <characters> <characters>-choice-1 <characters>': {
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <characters>',
+                '<maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <characters> <characters>-choice-1 <characters>',
+            },
+            '<content> <content>-choice-1 <nuls> <nuls>-choice-1 <nuls> <nuls>-choice-1 <NUL>': {
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls>',
+                '<maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls> <nuls>-choice-1 <NUL>',
+            },
+            '<content> <content>-choice-1 <nuls> <nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls>': {
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls>',
+                '<maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls>',
+            },
+            '<entries> <entries>-choice-1 <entry> <entry>-choice-1 <content> <content>-choice-1 <characters>': {
+                '<entries> <entries>-choice-1 <entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters>',
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+            },
+            '<entries> <entries>-choice-1 <entry> <entry>-choice-1 <content> <content>-choice-1 <nuls>': {
+                '<entries> <entries>-choice-1 <entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls>',
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+            },
+            '<entries> <entries>-choice-2 <entry> <entry>-choice-1 <content> <content>-choice-1 <characters>': {
+                '<entries> <entries>-choice-2 <entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters>',
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+            },
+            '<entries> <entries>-choice-2 <entry> <entry>-choice-1 <content> <content>-choice-1 <nuls>': {
+                '<entries> <entries>-choice-2 <entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls>',
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+            },
+            '<entry> <entry>-choice-1 <content> <content>-choice-1 <characters> <characters>-choice-1 <character>': {
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <character>',
+            },
+            '<entry> <entry>-choice-1 <content> <content>-choice-1 <characters> <characters>-choice-1 <characters>': {
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters>',
+                '<content> <content>-choice-1 <maybe_characters> <maybe_characters>-choice-1 <characters> <characters>-choice-1 <characters>',
+            },
+            '<entry> <entry>-choice-1 <content> <content>-choice-1 <nuls> <nuls>-choice-1 <NUL>': {
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <NUL>',
+            },
+            '<entry> <entry>-choice-1 <content> <content>-choice-1 <nuls> <nuls>-choice-1 <nuls>': {
+                '<entry> <entry>-choice-1 <content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls>',
+                '<content> <content>-choice-1 <maybe_nuls> <maybe_nuls>-choice-1 <nuls> <nuls>-choice-1 <nuls>',
+            },
+        }
+
         for job in self.jobs_and_generators:
             sids = tuple(get_all_sids(job, self.db_file))
             if self.num_sessions >= 0:
@@ -344,9 +422,25 @@ class Evaluator:
                     cur.execute(
                         "SELECT paths FROM inputs NATURAL JOIN kpaths WHERE testId = ? AND sid = ? AND k = ?",
                         (job, sid, k))
+
                     paths: Set[str] = set()
                     for row in cur:
-                        paths.update(set(next(ijson.items(row[0], ''))))
+                        for path in next(ijson.items(row[0].encode('utf-8'), '')):
+                            if path not in all_kpaths[k]:
+                                if path in tar_kpath_fix_map:
+                                    for fixed_path in tar_kpath_fix_map[path]:
+                                        assert fixed_path in all_kpaths[k], \
+                                            f'Fixed path {fixed_path} is broken'
+                                        paths.add(fixed_path)
+                                else:
+                                    print(
+                                        f'For {job}, session {sid}, k={k}, found a covered path that is not ' +
+                                        f'in the list of all covered paths (SKIPPING): "{path}"',
+                                        file=sys.stderr)
+
+                                continue
+
+                            paths.add(path)
 
                     if self.do_print_missing_kpaths:
                         missing_paths = all_kpaths[k].difference(paths)
@@ -358,10 +452,6 @@ class Evaluator:
 
                     diversity_by_k[k] = len(paths) / len(all_kpaths[k])
 
-                    assert all(path in all_kpaths[k] for path in paths), \
-                        f'For {job}, session {sid}, k={k}, found covered paths that are not ' \
-                        f'in the list of all covered paths.'
-
                 diversity_by_sid[sid] = sum(diversity_by_k.values()) / len(diversity_by_k)
 
             diversity[job] = sum(diversity_by_sid.values()) / len(diversity_by_sid)
@@ -374,7 +464,7 @@ class Evaluator:
 
             if valid_inputs:
                 def get_input_length(row: Sequence[str]) -> int:
-                    return len(tree_to_string(next(ijson.items(row[0], ''))))
+                    return len(tree_to_string(next(ijson.items(row[0].encode('utf-8'), ''))))
 
                 with pmp.ProcessingPool(processes=pmp.cpu_count()) as pool:
                     input_lengths = pool.map(get_input_length, valid_inputs)
@@ -382,7 +472,7 @@ class Evaluator:
                 # input_lengths = []
                 # for row in cur:
                 #     input_lengths.append(len(str(isla.derivation_tree.DerivationTree.from_parse_tree(
-                #         next(ijson.items(row[0], ''))))))
+                #         next(ijson.items(row[0].encode('utf-8'), ''))))))
 
                 avg_inp_length[job] = statistics.mean(input_lengths)
                 median_inp_length[job] = statistics.median(input_lengths)
@@ -647,11 +737,11 @@ def get_inputs_from_db(
 
     if convert_to_derivation_tree:
         inputs: Dict[int, isla.derivation_tree.DerivationTree] = {
-            row[0]: isla.derivation_tree.DerivationTree.from_parse_tree(next(ijson.items(row[1], '')))
+            row[0]: isla.derivation_tree.DerivationTree.from_parse_tree(next(ijson.items(row[1].encode('utf-8'), '')))
             for row in rows}
     else:
-        inputs: Dict[int, isla.derivation_tree.DerivationTree] = {
-            row[0]: next(ijson.items(row[1], ''))
+        inputs: Dict[int, ParseTree] = {
+            row[0]: next(ijson.items(row[1].encode('utf-8'), ''))
             for row in rows}
 
     return inputs
@@ -742,19 +832,36 @@ def evaluate_kpaths(
 
     def compute_k_paths(inp: ParseTree) -> List[str]:
         try:
+            assert isinstance(inp, list), f'Expected a list, got: {type(inp).__name__}'
+            assert len(inp) == 2, f'Expected a (node, children) pair, but tree has {len(inp)} items'
+            assert isinstance(inp[0], str), f'Node label should be a str, got {type(inp[0]).__name__}'
+            assert inp[1] is None or isinstance(inp[1], list), \
+                f'Children field should be None or list, is: {type(inp[1]).__name__}'
             return [path_to_string(path) for path in graph.k_paths_in_tree(inp, k, include_terminals=False)]
         except SyntaxError as err:
             print(f"Could not compute {k}-paths, error: {err}")
             return []
 
-    covered_k_paths = {inp_id: compute_k_paths(inp) for inp_id, inp in inputs.items()}
+    # We evaluate inputs in bundles of `bundle_size` inputs at a time
+    bundle_size = 1
+    get2nd = lambda t: t[1]
 
-    # Insert into DB
-    con = sqlite3.connect(db_file)
-    with con:
-        con.executemany(
-            "INSERT INTO kpaths(inpId, k, paths) VALUES (?, ?, ?)",
-            [(inp_id, k, json.dumps(kpaths)) for inp_id, kpaths in covered_k_paths.items()])
-    con.close()
+    input_bundle: List[Tuple[int, Tuple[int, ParseTree]]]
+    for input_bundle in map(list, map(
+            get2nd,
+            itertools.groupby(enumerate(inputs.items()),
+                              lambda p: p[0] // bundle_size))):
+        print(f'Evaluating {k}-paths for session {sid} of {test_id}: '
+              f'{100 * input_bundle[0][0] / len(inputs):.2f}% done '
+              f'({input_bundle[0][0]} of {len(inputs)} inputs)')
+        covered_k_paths = {inp_id: compute_k_paths(inp) for _, (inp_id, inp) in input_bundle}
+
+        # Insert into DB
+        con = sqlite3.connect(db_file)
+        with con:
+            con.executemany(
+                "INSERT INTO kpaths(inpId, k, paths) VALUES (?, ?, ?)",
+                [(inp_id, k, json.dumps(kpaths)) for inp_id, kpaths in covered_k_paths.items()])
+        con.close()
 
     print(f"[{strtime()}] DONE Evaluating {k}-paths for session {sid} of {test_id}")
